use core::panic;
use std::collections::{HashMap, HashSet};

#[derive(Clone, Copy, Hash, PartialEq, Eq)]
enum Orientation {
    North,
    East,
    South,
    West,
}

impl Orientation {
    fn turn(&mut self, direction: &str) {
        match direction {
            "R" => match self {
                Orientation::North => *self = Self::East,
                Orientation::East => *self = Self::South,
                Orientation::South => *self = Self::West,
                Orientation::West => *self = Self::North,
            },
            "L" => match self {
                Orientation::North => *self = Self::West,
                Orientation::West => *self = Self::South,
                Orientation::South => *self = Self::East,
                Orientation::East => *self = Self::North,
            },
            _ => panic!("Wrong direction. Only R and L allowed"),
        }
    }
}

pub fn process_part1(input: &str) -> i32 {
    let mut orientation = Orientation::North;
    let mut movements = HashMap::new();
    let mut input = input.to_string();
    input.pop();
    for movement in input.split(", ") {
        let (direction, blocks) = movement.split_at(1);
        orientation.turn(direction);
        let blocks = blocks.parse::<i32>().unwrap();
        movements
            .entry(orientation)
            .and_modify(|distance| *distance += blocks)
            .or_insert(blocks);
    }
    let up = movements.get(&Orientation::North).unwrap_or(&0);
    let down = movements.get(&Orientation::South).unwrap_or(&0);
    let left = movements.get(&Orientation::West).unwrap_or(&0);
    let right = movements.get(&Orientation::East).unwrap_or(&0);
    (up - down).abs() + (left - right).abs()
}

pub fn process_part2(input: &str) -> i32 {
    let mut orientation = Orientation::North;
    let mut visited_coords = HashSet::new();
    let mut horizontal: i32 = 0;
    let mut vertical: i32 = 0;
    let mut input = input.to_string();
    input.pop();
    for movement in input.split(", ") {
        let (direction, blocks) = movement.split_at(1);
        orientation.turn(direction);
        let blocks = blocks.parse::<i32>().unwrap();
        for _block in 1..=blocks {
            match orientation {
                Orientation::North => vertical += 1,
                Orientation::South => vertical -= 1,
                Orientation::East => horizontal += 1,
                Orientation::West => horizontal -= 1,
            }
            if visited_coords.contains(&(horizontal, vertical)) {
                return horizontal.abs() + vertical.abs();
            } else {
                visited_coords.insert((horizontal, vertical));
            }
        }
    }
    panic!("Did not find easter bunny hq");
}

#[cfg(test)]
mod tests {
    use super::*;

    const INPUT1: &str = "R2, L3 ";
    const INPUT2: &str = "R2, R2, R2 ";
    const INPUT3: &str = "R5, L5, R5, R3 ";
    const INPUT4: &str = "R8, R4, R4, R8 ";

    #[test]
    fn part1() {
        let result = process_part1(INPUT1);
        assert_eq!(result, 5);
        let result = process_part1(INPUT2);
        assert_eq!(result, 2);
        let result = process_part1(INPUT3);
        assert_eq!(result, 12);
    }

    #[test]
    fn part2() {
        let result = process_part2(INPUT4);
        assert_eq!(result, 4);
    }
}
